#include "preferences.h"
#include "profiles.h"
#include "version.h"
#include "vimpal.h"

#include <QCloseEvent>
#include <QComboBox>
#include <QFileSystemModel>
#include <QMenu>
#include <QMessageBox>
#include <QProcess>
#include <QSettings>
#include <QStatusBar>
#include <QTimer>
#include <QToolBar>
#include <QTreeView>
#include <QPushButton>
#include <QInputDialog>

#include <QtDebug>


VimPal::VimPal(QWidget* parent, Qt::WindowFlags f)
    :QWidget(parent, f)
{
    setObjectName("VimPal");
    setWindowTitle(tr("VIM Pal"));
    setWindowIcon(QIcon(":/img/icon.png"));

    QVBoxLayout* vl = new QVBoxLayout(this);

    vl->setMargin(0);
    vl->setSpacing(0);

    mRootMenu = new QMenu(this);

    connect(mRootMenu, SIGNAL(aboutToShow()), this, SLOT(updateRootMenu()));
    connect(mRootMenu, SIGNAL(triggered(QAction*)), this, SLOT(changeRoot(QAction*)));

    // Toolbar
    {
        QToolBar* tb = new QToolBar(this);

        vl->addWidget(tb);

        tb->setIconSize(QSize(22, 22));

        QToolButton* b = new QToolButton(this);
        QMenu* m = new QMenu(this);

        m->addAction(tr("&Preferences..."), this, SLOT(editPreferences()));
        m->addSeparator();
        m->addAction(tr("&About VIM Pal..."), this, SLOT(showAbout()));
        m->addSeparator();
        m->addAction(tr("&Quit"), this, SLOT(close()))->setShortcut(QKeySequence("Ctrl+Q"));

        b->setMenu(m);
        b->setIcon(QIcon(":/img/icon.png"));
        b->setPopupMode(QToolButton::InstantPopup);

        tb->addWidget(b);
        tb->addSeparator();
        (mGoBack = tb->addAction(QIcon::fromTheme("go-previous", QIcon(":/img/go-previous.png")), tr("Go back"), this, SLOT(goBack())))->setShortcut(QKeySequence("Alt+Left"));
        (mGoForward = tb->addAction(QIcon::fromTheme("go-next", QIcon(":/img/go-next.png")), tr("Go forward"), this, SLOT(goForward())))->setShortcut(QKeySequence("Alt+Right"));

        b = new QToolButton(this);

        b->setIcon(QIcon::fromTheme("go-up", QIcon(":/img/go-up.png")));
        b->setToolTip(tr("Go up"));
        b->setShortcut(QKeySequence("Alt+Up"));
        b->setPopupMode(QToolButton::MenuButtonPopup);
        b->setMenu(mRootMenu);

        connect(b, SIGNAL(clicked()), this, SLOT(goUp()));

        tb->addWidget(b);

        tb->addAction(QIcon::fromTheme("go-home", QIcon(":/img/go-home.png")), tr("Go home"), this, SLOT(goHome()))->setShortcut(QKeySequence("Alt+Home"));
    }

    // Tree
    mDirView = new QTreeView(this);
    mDirModel = new QFileSystemModel(this);

    connect(mDirView, SIGNAL(clicked(QModelIndex)), this, SLOT(fileClick(QModelIndex)));
    connect(mDirView, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(fileDblClick(QModelIndex)));
    connect(mDirView, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(contextMenu(QPoint)));

    vl->addWidget(mDirView);

    mDirView->setModel(mDirModel);
    mDirView->hideColumn(1);
    mDirView->hideColumn(2);
    mDirView->hideColumn(3);
    mDirView->header()->setResizeMode(QHeaderView::ResizeToContents);
    mDirView->setHeaderHidden(true);
    mDirView->setContextMenuPolicy(Qt::CustomContextMenu);

    mDirModel->sort(0);

    // Status bar
    {
        QStatusBar* sb = new QStatusBar(this);

        vl->addWidget(sb);

        mProfileBox = new QComboBox(this);
        QToolButton* tb = new QToolButton(this);

        sb->addWidget(mProfileBox, 10);
        sb->addWidget(tb);

        tb->setText("...");

        connect(tb, SIGNAL(clicked()), this, SLOT(editProfiles()));
        connect(mProfileBox, SIGNAL(currentIndexChanged(int)), this, SLOT(profileChanged(int)));
    }

    {
        QSettings set;

        restoreGeometry(set.value("window/geometry").toByteArray());

        mSingleClickOpen = set.value("single_click", false).toBool();
        mChangeDir = set.value("change_dir", true).toBool();

#ifdef Q_WS_WIN
        mVimCommand = set.value("vim_command", "C:/Program Files/Vim/vim73/gvim.exe").toString();
#else
        mVimCommand = set.value("vim_command", "gvim").toString();
#endif  // Q_WS_WIN

        mOpenNewTabs = set.value("new_tabs", true).toBool();
        mCurrentProfile = set.value("last_profile", 0).toInt();
        mDarkBackground = set.value("dark_background", false).toBool();
        mShowHidden = set.value("show_hidden", false).toBool();

        const int c = set.beginReadArray("profiles");

        if (!c)
        {
            mProfiles.append(Profile(tr("Default"), "GVIM", QDir::homePath()));

            set.endArray();
            set.beginWriteArray("profiles", 1);
            set.setArrayIndex(0);
            set.setValue("name", tr("Default"));
            set.setValue("server_name", "VIM");
        }
        else
        {
            for (int i=0; i<c; i++)
            {
                set.setArrayIndex(i);

                mProfiles.append(Profile(set.value("name").toString(), set.value("server_name").toString(), set.value("root_dir").toString()));
            }
        }

        set.endArray();

        mDirModel->setRootPath(mProfiles[mCurrentProfile].Dir);
        mDirView->setRootIndex(mDirModel->index(mProfiles[mCurrentProfile].Dir));
    }

    updateProfiles();
    updateColors();
    clearHistory();

    mDirModel->setFilter(mShowHidden ? (QDir::AllEntries | QDir::AllDirs | QDir::NoDotAndDotDot | QDir::Hidden) :
                                       (QDir::AllEntries | QDir::AllDirs | QDir::NoDotAndDotDot));
}

/* --------------------------------------------------------------------- */

void VimPal::closeEvent(QCloseEvent* e)
{
    QSettings set;

    set.setValue("window/geometry", saveGeometry());

    e->accept();
}

/* --------------------------------------------------------------------- */

void VimPal::goHome()
{
    selectDir(mDirModel->index(QDir::homePath()));
}

void VimPal::goUp()
{
    if (mDirView->rootIndex().parent().isValid())
    {
        selectDir(mDirView->rootIndex().parent());
    }
}

void VimPal::goBack()
{
    if (!mHistory.isEmpty())
    {
        const QString p = mHistory.pop();
        mForwardHistory.push(mDirModel->rootPath());

        mGoForward->setEnabled(true);

        if (mHistory.isEmpty())
        {
            mGoBack->setDisabled(true);
        }

        selectDir(mDirModel->index(p), true);
    }
}

void VimPal::goForward()
{
    if (!mForwardHistory.isEmpty())
    {
        const QString p = mForwardHistory.pop();
        mHistory.push(mDirModel->rootPath());

        mGoBack->setEnabled(true);

        if (mForwardHistory.isEmpty())
        {
            mGoForward->setDisabled(true);
        }

        selectDir(mDirModel->index(p), true);
    }
}

void VimPal::changeRoot(QAction *action)
{
    if (!action)
    {
        return;
    }

    const QString root = action->property("path").toString();

    if (!root.isEmpty())
    {
        selectDir(mDirModel->index(root));
    }
}

void VimPal::updateRootMenu()
{
    mRootMenu->clear();

    foreach (const QFileInfo& drive, QDir::drives())
    {
        mRootMenu->addAction(drive.absoluteFilePath())->setProperty("path", drive.absoluteFilePath());
    }
}

void VimPal::fileClick(const QModelIndex &idx)
{
    if (mSingleClickOpen)
    {
        if (mDirModel->isDir(idx))
        {
            if (mChangeDir)
            {
                vimCmd(QString("--remote-send \"<C-\\><C-N>:cd %1<CR>\"").arg(mDirModel->filePath(idx)), false);
            }
        }
        else
        {
            selectFile(idx);
        }
    }
}

void VimPal::fileDblClick(const QModelIndex &idx)
{
    if (mDirModel->isDir(idx))
    {
        selectDir(idx);
    }
    else if (!mSingleClickOpen)
    {
        selectFile(idx);
    }
}

void VimPal::contextMenu(const QPoint &pos)
{
    const QModelIndex idx = mDirView->indexAt(pos);

    QMenu m(this), newMenu(tr("&New"), this);
    QMap<QAction*,int> actions;

    if (idx.isValid())
    {
        mDirView->selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);

        if (mDirModel->isDir(idx))
        {
            actions.insert(m.addAction(tr("&Set as current in Vim")), 1);
            actions.insert(m.addAction(tr("&Set as root")), 2);
        }
        else
        {
            actions.insert(m.addAction(tr("&Open in Vim")), 3);
        }

        m.addSeparator();

        actions.insert(m.addAction(tr("&Rename...")), 4);
        actions.insert(m.addAction(tr("&Delete...")), 5);
    }
    else
    {
        mDirView->selectionModel()->clear();
    }

    actions.insert(newMenu.addAction(tr("&File...")), 6);
    actions.insert(newMenu.addAction(tr("&Directory...")), 7);

    m.addMenu(&newMenu);

    QAction* a = m.exec(mDirView->mapToGlobal(pos) + QPoint(1, 1));

    switch (actions.value(a, -1))
    {
        case 1:
            vimCmd(QString("--remote-send \"<C-\\><C-N>:cd %1<CR>\"").arg(mDirModel->filePath(idx)), false);
            break;

        case 2:
            selectDir(idx);
            break;

        case 3:
            selectFile(idx);
            break;

        case 4:
            renameFile(idx);
            break;

        case 5:
            deleteFile(idx);
            break;

        case 6:
            newFile(idx);
            break;

        case 7:
            newDirectory(idx);
            break;
    }
}

void VimPal::editPreferences()
{
    Preferences dlg(this);

    dlg.singleClickOpen->setChecked(mSingleClickOpen);
    dlg.changeDir->setChecked(mChangeDir);
    dlg.vimCommand->setText(mVimCommand);
    dlg.newTabs->setChecked(mOpenNewTabs);
    dlg.darkBackground->setChecked(mDarkBackground);
    dlg.showHidden->setChecked(mShowHidden);

    if (dlg.exec() == QDialog::Accepted)
    {
        mSingleClickOpen = dlg.singleClickOpen->isChecked();
        mChangeDir = dlg.changeDir->isChecked();
        mVimCommand = dlg.vimCommand->text();
        mOpenNewTabs = dlg.newTabs->isChecked();
        mDarkBackground = dlg.darkBackground->isChecked();
        mShowHidden = dlg.showHidden->isChecked();

        QSettings set;

        set.setValue("single_click", mSingleClickOpen);
        set.setValue("change_dir", mChangeDir);
        set.setValue("vim_command", mVimCommand);
        set.setValue("new_tabs", mOpenNewTabs);
        set.setValue("dark_background", mDarkBackground);
        set.setValue("show_hidden", mShowHidden);

        updateColors();
        mDirModel->setFilter(mShowHidden ? (QDir::AllEntries | QDir::AllDirs | QDir::NoDotAndDotDot | QDir::Hidden)
                                         : (QDir::AllEntries | QDir::AllDirs | QDir::NoDotAndDotDot));
    }
}

void VimPal::editProfiles()
{
    Profiles dlg(this);

    dlg.setProfiles(mProfiles);

    if (dlg.exec() == QDialog::Accepted)
    {
        mProfiles = dlg.getProfiles();
        mCurrentProfile = dlg.getNewIndex(mCurrentProfile);

        QSettings set;

        set.beginWriteArray("profiles", mProfiles.count());

        for (int i=0, c=mProfiles.count(); i<c; i++)
        {
            set.setArrayIndex(i);

            set.setValue("name", mProfiles.at(i).Title);
            set.setValue("server_name", mProfiles.at(i).Server);
            set.setValue("root_dir", mProfiles.at(i).Dir);
        }

        set.endArray();

        if (mCurrentProfile == -1)
        {
            mCurrentProfile = 0;

            if (!mProfiles[mCurrentProfile].Dir.isEmpty())
            {
                mDirModel->setRootPath(mProfiles[mCurrentProfile].Dir);
                mDirView->setRootIndex(mDirModel->index(mProfiles[mCurrentProfile].Dir));
            }

            clearHistory();
        }

        set.setValue("last_profile", mCurrentProfile);

        updateProfiles();
    }
}

void VimPal::profileChanged(int prof)
{
    if (mCurrentProfile != prof)
    {
        mCurrentProfile = prof;

        if (!mProfiles[mCurrentProfile].Dir.isEmpty())
        {
            mDirModel->setRootPath(mProfiles[mCurrentProfile].Dir);
            mDirView->setRootIndex(mDirModel->index(mProfiles[mCurrentProfile].Dir));
        }
        else
        {
            const QString dir = mDirModel->fileInfo(mDirView->rootIndex()).absoluteFilePath();
            QSettings set;

            mProfiles[mCurrentProfile].Dir = dir;

            set.setValue(QString("profiles/%1/root_dir").arg(mCurrentProfile + 1), dir);
        }

        QSettings set;

        set.setValue("last_profile", mCurrentProfile);

        clearHistory();
    }
}

void VimPal::showAbout()
{
    QMessageBox::about(this, tr("VIM Pal"),
        tr("<big><strong>VIM Pal</strong></big><br /><em>Version %1</em><br /><br />&copy; %2").arg(VERSION).arg(COPYRIGHT));
}

/* --------------------------------------------------------------------- */

bool VimPal::vimCmd(const QString &cmd, bool startNew)
{
    const QString prof = mProfiles[mCurrentProfile].Server;

    bool ok = true, haveServer = vimServerExists(prof);

    if (!haveServer && startNew)
    {
        QProcess* p = new QProcess(this);

        connect(p, SIGNAL(finished(int)), p, SLOT(deleteLater()));

        p->start(QString("\"%1\" --servername \"%2\"").arg(mVimCommand).arg(prof));

        if (p->waitForStarted())
        {
            int checks = 6;

            while (--checks)
            {
                QEventLoop loop;
                QTimer::singleShot(500, &loop, SLOT(quit()));
                loop.exec();

                if (vimServerExists(prof))
                {
                    haveServer = true;

                    break;
                }
            }
        }
        else
        {
            ok = false;
        }
    }

    if (haveServer)
    {
        QProcess* p = new QProcess(this);

        connect(p, SIGNAL(finished(int)), p, SLOT(deleteLater()));

        p->start(QString("\"%1\" --servername \"%2\" %3").arg(mVimCommand).arg(prof).arg(cmd));

        if (!p->waitForStarted())
        {
            ok = false;
        }
    }
    else if (startNew)
    {
        ok = false;
    }

    if (!ok)
    {
        QMessageBox::critical(this, tr("Error"),
            tr("Cannot execute VIM command. Please check your preferences and set the correct path to VIM executable."));
    }

    return ok;
}

bool VimPal::vimServerExists(const QString &name)
{
    QProcess* p = new QProcess(this);
    bool ok = false;

#ifdef Q_WS_WIN
    // XXX: funny stuff on windows... as always
    {
        QString vcmd = mVimCommand;

        vcmd.replace("gvim.exe", "vim.exe", Qt::CaseInsensitive);

        p->start(QString("\"%1\" --serverlist").arg(vcmd));
    }
#else
    p->start(QString("\"%1\" --serverlist").arg(mVimCommand));
#endif  // Q_WS_WIN

    if (p->waitForFinished())
    {
        while (!p->atEnd())
        {
            if (p->readLine().trimmed() == name)
            {
                ok = true;

                break;
            }
        }
    }

    p->deleteLater();

    return ok;
}

void VimPal::selectFile(const QModelIndex &index)
{
    if (mOpenNewTabs)
    {
        if (!vimCmd(QString("--remote-tab \"%1\"").arg(mDirModel->filePath(index)), true))
        {
            return;
        }
    }
    else
    {
        if (!vimCmd(QString("--remote \"%1\"").arg(mDirModel->filePath(index)), true))
        {
            return;
        }
    }

    if (mChangeDir)
    {
        vimCmd(QString("--remote-send \"<C-\\><C-N>:cd %1<CR>\"").arg(mDirModel->fileInfo(index).dir().absolutePath()), false);
    }
}

void VimPal::selectDir(const QModelIndex &idx, bool noHistory)
{
    const QString dir = mDirModel->fileInfo(idx).absoluteFilePath();

    if (dir == mDirModel->rootPath())
    {
        return;
    }

    if (!noHistory)
    {
        pushHistory(mDirModel->rootPath());
    }

    mDirModel->setRootPath(dir);
    mDirView->setRootIndex(idx);

    mProfiles[mCurrentProfile].Dir = dir;

    QSettings set;

    set.setValue(QString("profiles/%1/root_dir").arg(mCurrentProfile + 1), dir);
}

void VimPal::deleteFile(const QModelIndex &idx)
{
    const QFileInfo info = mDirModel->fileInfo(idx);

    QMessageBox box(QMessageBox::Warning, tr("Confirm delete"),
                    tr("Are you sure you want to delete %1 <b>%2</b>?").arg(info.isDir() ? tr("directory") : tr("file")).arg(info.fileName()),
                    QMessageBox::Yes | QMessageBox::No, this);

    box.button(QMessageBox::Yes)->setText(tr("&Yes"));
    box.button(QMessageBox::No)->setText(tr("&No"));

    if (box.exec() == QMessageBox::Yes)
    {
        if (info.isDir())
        {
            if (!mDirModel->rmdir(idx))
            {
                QMessageBox::critical(this, tr("Error"), tr("Unable to remove directory."));
            }
        }
        else
        {
            QFile f(info.absoluteFilePath());

            if (!f.remove())
            {
                QMessageBox::critical(this, tr("Error"), tr("Unable to remove file: %1").arg(f.errorString()));
            }
        }
    }
}

void VimPal::renameFile(const QModelIndex &idx)
{
    const QFileInfo info = mDirModel->fileInfo(idx);

    QInputDialog dlg(this);

    dlg.setOkButtonText(tr("&OK"));
    dlg.setCancelButtonText(tr("&Cancel"));
    dlg.setWindowTitle(info.isDir() ? tr("Rename directory") : tr("Rename file"));
    dlg.setLabelText(info.isDir() ? tr("Enter new directory name:") : tr("Enter new file name:"));
    dlg.setTextValue(info.fileName());

    if (dlg.exec() == QDialog::Accepted)
    {
        QFile f(info.absoluteFilePath());

        if (!f.rename(info.absoluteDir().absoluteFilePath(dlg.textValue())))
        {
            QMessageBox::critical(this, tr("Error"), tr("Unable to rename: %1").arg(f.errorString()));
        }
    }
}

void VimPal::newFile(const QModelIndex &idx)
{
    QInputDialog dlg(this);

    dlg.setOkButtonText(tr("&OK"));
    dlg.setCancelButtonText(tr("&Cancel"));
    dlg.setWindowTitle(tr("Create new file"));
    dlg.setLabelText(tr("Enter the name for new file:"));

    if (dlg.exec() == QDialog::Accepted)
    {
        const QString name = getDir(idx) + "/" + dlg.textValue();

        if (QFile::exists(name))
        {
            QMessageBox::critical(this, tr("Error"), tr("File already exists."));
        }
        else
        {
            QFile f(name);

            if (f.open(QIODevice::WriteOnly))
            {
                f.close();
            }
            else
            {
                QMessageBox::critical(this, tr("Error"), tr("Unable to create file: %1").arg(f.errorString()));
            }
        }
    }
}

void VimPal::newDirectory(const QModelIndex &idx)
{
    QInputDialog dlg(this);

    dlg.setOkButtonText(tr("&OK"));
    dlg.setCancelButtonText(tr("&Cancel"));
    dlg.setWindowTitle(tr("Create new directory"));
    dlg.setLabelText(tr("Enter the name for new directory:"));

    if (dlg.exec() == QDialog::Accepted)
    {
        if (!mDirModel->mkdir(mDirModel->index(getDir(idx)), dlg.textValue()).isValid())
        {
            QMessageBox::critical(this, tr("Error"), tr("Unable to create directory."));
        }
    }
}

void VimPal::updateProfiles()
{
    mProfileBox->blockSignals(true);
    mProfileBox->clear();

    foreach (const Profile& p, mProfiles)
    {
        mProfileBox->addItem(p.Title);
    }

    mProfileBox->setCurrentIndex(mCurrentProfile);
    mProfileBox->blockSignals(false);
}

void VimPal::updateColors()
{
    if (mDarkBackground)
    {
        mDirView->setStyleSheet("background-color: #222; color: #EEE;");
    }
    else
    {
        mDirView->setStyleSheet("");
    }
}

void VimPal::clearHistory()
{
    mGoBack->setDisabled(true);
    mGoForward->setDisabled(true);
    mHistory.clear();
    mForwardHistory.clear();
}

void VimPal::pushHistory(const QString &path)
{
    if (mHistory.isEmpty() || mHistory.top() != path)
    {
        mGoBack->setEnabled(true);
        mHistory.push(path);

        mGoForward->setDisabled(true);
        mForwardHistory.clear();
    }
}

QString VimPal::getDir(const QModelIndex &idx)
{
    const QFileInfo info = mDirModel->fileInfo(idx.isValid() ? idx : mDirView->rootIndex());

    return info.isDir() ? info.absoluteFilePath() : info.absolutePath();
}
